#include "stdafx.h"
#include "UIC_StateTracker.h"

#include "../Launcher/UIC_SharedMemory.h"
#include "Interfaces/UIC_IDirect3DTexture9.h"

using namespace std;

UIC_StateTracker::UIC_StateTracker()
	: myFrameCounter(0)
	, myTextureBindCounter(0)
	, mySharedMemory(UIC_GetSharedMemory())
	, myLastFrameFlags(0)
	, myCurrentFrameFlags(0)
	, myIsDevMode(false)
	, myCurrentSelection(0)
	, myCombinedFlagsWithTemp(UIC_SharedMemory::Always)
{
	if (UIC_SharedMemory* shmem = mySharedMemory)
	{
		myIsDevMode = shmem->myIsDevMode != 0;
		myCombinedFlagsWithTemp = shmem->myFlags;
	}
}

UIC_StateTracker::~UIC_StateTracker()
{
}

bool UIC_StateTracker::Bind(UIC_IDirect3DTexture9* aTexture)
{
	aTexture->myLastBindFrame = myFrameCounter;
	aTexture->myLastBindCounter = myTextureBindCounter++;
	myCurrentFrameFlags |= aTexture->mySetOnBindFlags;

	if (myCurrentSelection == aTexture)
		return myFrameCounter & 1;
	return true;
}

void UIC_StateTracker::NewFrame()
{
	++myFrameCounter;
	myTextureBindCounter = 0;
	myCombinedFlagsWithTemp = myLastFrameFlags = myCurrentFrameFlags;
	myCurrentFrameFlags = 0;

	if (UIC_SharedMemory* shmem = mySharedMemory)
	{
		UIC_SharedMemory::Command command = shmem->myCommand;
		shmem->myCommand = UIC_SharedMemory::Command::Nothing;
		shmem->myFrameCounter = myFrameCounter;

		shmem->myFlagsFromGame = myLastFrameFlags;
		myCombinedFlagsWithTemp = shmem->UpdateFlags();

		if (myIsDevMode)
		{
			bool ok = false;
			switch (command)
			{
			case UIC_SharedMemory::ClearSelection:
				myCurrentSelection = 0;
				ok = true;
				break;
			case UIC_SharedMemory::SelectLastTexture:
				ok = SelectLast();
				break;
			case UIC_SharedMemory::SelectPreviousTexture:
				ok = StepSelection(-1);
				break;
			case UIC_SharedMemory::SelectNextTexture:
				ok = StepSelection(1);
				break;
			case UIC_SharedMemory::SaveCurrentTextureInfo:
				ok = SaveSelected();
				break;
			case UIC_SharedMemory::Nothing:
				goto nobeep;
			}
			if (ok)
				Beep(1000, 50);
			else
				Beep(100, 100);

			nobeep:;
		}
	}
}

bool UIC_StateTracker::NeedsInit(unsigned aWidth, unsigned aHeight, unsigned aFormat, unsigned aLevels)
{
	if (aLevels != 1 || aFormat != D3DFMT_DXT5)
		return false;

	if (myIsDevMode)
		return true;

	if (mySharedMemory)
	{
		for (unsigned i=0, e=mySharedMemory->myNumKnownSizes; i!=e; ++i)
		{
			UIC_SharedMemory::KnownSize size = mySharedMemory->myKnownSizes[i];
			if (size.myWidth == aWidth && size.myHeight == aHeight)
				return true;
		}
	}

	return false;
}

void UIC_StateTracker::AddTexture(UIC_IDirect3DTexture9* aTexture)
{
	if (mySharedMemory)
	{
		unsigned w = aTexture->myWidth;
		unsigned h = aTexture->myHeight;
		unsigned crc = aTexture->myCRC32;
		for (unsigned i=0, e=mySharedMemory->myNumKnownTextures; i!=e; ++i)
		{
			const UIC_SharedMemory::TextureInfo& info = mySharedMemory->myKnownTextures[i];
			if (info.myWidth == w && info.myHeight == h && info.myCRC32 == crc)
			{
				aTexture->mySetOnBindFlags = info.mySetOnBindFlags;
				aTexture->mySuppressDrawingWhenActiveFlags = info.mySuppressDrawingWhenActiveFlags;
				aTexture->mySetTempFlags = info.mySetTempFlags;
				aTexture->myClearTempFlags = info.myClearTempFlags;
				break;
			}
		}
	}

	if (!myIsDevMode)
		return;

	lock_guard<mutex> lock(myTexturesMutex);
	aTexture->myIsAddedToList = true;
	myTextures.push_back(aTexture);
}

void UIC_StateTracker::RemoveTexture(UIC_IDirect3DTexture9* aTexture)
{
	if (!myIsDevMode)
		return;

	if (myCurrentSelection == aTexture)
		myCurrentSelection = 0;

	lock_guard<mutex> lock(myTexturesMutex);
	auto it = find(myTextures.begin(), myTextures.end(), aTexture);
	if (it != myTextures.end())
	{
		auto last = myTextures.end()-1;
		if (it != last)
			iter_swap(it, last);
		myTextures.pop_back();
	}
}

bool UIC_StateTracker::StepSelection(int aDelta)
{
	if (!myIsDevMode)
		return false;

	lock_guard<mutex> lock(myTexturesMutex);
	auto it = find(myTextures.begin(), myTextures.end(), myCurrentSelection);
	if (it == myTextures.end())
	{
		if (myTextures.size())
			myCurrentSelection = myTextures.back();
		else
			myCurrentSelection = 0;
		return false;
	}
	else
	{
		int index = it - myTextures.begin();
		int oldindex = index;
		index += aDelta;
		index = max(index, 0);
		index = min(index, int(myTextures.size()));
		if (oldindex != index)
		{
			if (index == myTextures.size())
			{
				myCurrentSelection = 0;
				return false;
			}
			else
			{
				myCurrentSelection = myTextures[index];
				if (myCurrentSelection->myLastBindFrame < myFrameCounter - 5)
					return false;
				return true;
			}
		}
		return false;
	}
}

bool UIC_StateTracker::SelectLast()
{
	if (!myIsDevMode)
		return false;

	lock_guard<mutex> lock(myTexturesMutex);
	if (myTextures.size())
	{
		sort(myTextures.begin(), myTextures.end(), [](UIC_IDirect3DTexture9* l, UIC_IDirect3DTexture9* r){
			if (l->myLastBindFrame == r->myLastBindFrame)
				return l->myLastBindCounter < r->myLastBindCounter;
			return l->myLastBindFrame < r->myLastBindFrame;
		});

		myCurrentSelection = myTextures.back();
		return true;
	}
	return false;
}

bool UIC_StateTracker::SaveSelected()
{
	if (!myIsDevMode)
		return false;

	lock_guard<mutex> lock(myTexturesMutex);
	if (!myCurrentSelection)
		return false;

	UIC_IDirect3DTexture9* tex = myCurrentSelection;
	if (!tex->myCRC32)
		return false;

	FILE* f = 0;
	fopen_s(&f, "UIC Saved.txt", "a");
	if (!f)
		return false;

	fprintf(f, "%5i\t %5i\t\t%08X\t\t%X\t\t%X\t\t%X\t\t%X\n",
			tex->myWidth, tex->myHeight, tex->myCRC32,
			int(tex->mySetOnBindFlags), int(tex->mySuppressDrawingWhenActiveFlags), int(tex->mySetTempFlags), int(tex->myClearTempFlags));
	fclose(f);

	return true;
}
