#include "stdafx.h"
#include "..\\detours\\detours.h"

#include "UIC_SharedMemory.h"

#ifdef _DEBUG
#	define CONFIGSTRING "Debug"
#else
#	define CONFIGSTRING "Release"
#endif

#pragma comment(lib, "../" CONFIGSTRING "/detours.lib")

using namespace std;

void LoadSettings(UIC_SharedMemory* shmem, const char* aFilename, bool aAllowFailure = false);
void CreateConsole();
void StartHooked(HANDLE& aProcess, HWND& aWindow);
string FormatError(DWORD aError);

BOOL WaitWithMessageLoop( 
	HANDLE hHandleToWaitOn, 
	DWORD dwInitialTimeOutMilliseconds, 
	DWORD dwIterateTimeOutMilliseconds);

int _tmain(int argc, _TCHAR* argv[])
{
	ostream& log = cout;
	try
	{
		HANDLE singleInstance = CreateMutex(0, 0, "UIC_SingleInstanceMutex");
		if (GetLastError() == ERROR_ALREADY_EXISTS)
		{
			log << "Another instance is already running.";
			return 1;
		}

		if (FindWindowA("ArenaNet_Dx_Window_Class", "Guild Wars 2"))
			throw "Guild Wars 2 is already running. Please close the game and start UIC.exe to launch it.";

		char path[MAX_PATH] = {};
		GetModuleFileName(0, path, sizeof path);

		if (char* end = strrchr(path, '\\'))
			*end = 0;

#ifndef _DEBUG
		SetCurrentDirectory(path);
#endif

		UIC_SharedMemory* shmem = UIC_GetSharedMemory();
		if (!shmem)
			throw make_pair("Could not initialize shared memory", GetLastError());

		memset(shmem, 0, sizeof *shmem);
		shmem->myHasFocus = true;

		LoadSettings(shmem, "UIC Textures.txt");

#ifndef _DEBUG
		if (shmem->myIsDevMode)
			CreateConsole();
#else
		shmem->myIsDevMode = true;
#endif

		HANDLE process = 0;
		HWND window = 0;
		StartHooked(process, window);

		if (!window)
			shmem->myFlagsFromGame = UIC_SharedMemory::CombatPossible;

		WaitWithMessageLoop(process, 0, 50);

		return 0;
	}
	catch (const char* err)
	{
		log << err << endl;

		MessageBox(0, err, "Error", MB_ICONERROR|MB_OK);

		return 1;
	}
	catch (string err)
	{
		log << err << endl;

		MessageBox(0, err.c_str(), "Error", MB_ICONERROR|MB_OK);

		return 1;
	}
	catch (pair<const char*, DWORD> err)
	{
		string fmt = FormatError(err.second);
		log << err.first << ", " << fmt << endl;

		string str = string(err.first) + '\n' + fmt;
		MessageBox(0, str.c_str(), "Error", MB_ICONERROR|MB_OK);

		return 1;
	}
	catch (int err)
	{
		return err;
	}
}

void LoadSettings(UIC_SharedMemory* shmem, const char* aFilename, bool aAllowFailure)
{
	std::ifstream file(aFilename);
	if (!file.is_open())
	{
		if (aAllowFailure)
			return;
		throw string("Could not open ") + aFilename + ", please make sure UIC.exe and all related files are placed in the Guild Wars 2 folder.";
	}

	char line[512];
	while (file.getline(line, sizeof line))
	{
		int w, h, crc, flags, suppress = 0, set = 0, clear = 0;

		if (sscanf_s(line, "start %x", &flags) == 1)
			shmem->myFlagsFromInput = flags;
		else if (strncmp(line, "devmode", 7) == 0)
			shmem->myIsDevMode = true;
		else if (sscanf_s(line, "%i %i %x %x %x %x %x", &w, &h, &crc, &flags, &suppress, &set, &clear) >= 4)
		{
			if (shmem->myNumKnownTextures >= UIC_SharedMemory::MaxKnownTextures)
				throw "Too many textures recorded";

			UIC_SharedMemory::TextureInfo info(w, h, crc, flags, suppress, set, clear);
			shmem->myKnownTextures[shmem->myNumKnownTextures++] = info;

			for (int i=0, e=shmem->myNumKnownSizes; i!=e; ++i)
				if (shmem->myKnownSizes[i].myWidth == w && shmem->myKnownSizes[i].myHeight == h)
					goto size_found;

			{
				UIC_SharedMemory::KnownSize size = { w, h };
				shmem->myKnownSizes[shmem->myNumKnownSizes++] = size;
			}

size_found:	;
		}
	}

	shmem->UpdateFlags();
}

void CreateConsole()
{
	AllocConsole();

	HANDLE handle_out = GetStdHandle(STD_OUTPUT_HANDLE);
	int hCrt = _open_osfhandle((long) handle_out, _O_TEXT);
	FILE* hf_out = _fdopen(hCrt, "w");
	setvbuf(hf_out, NULL, _IONBF, 1);
	*stdout = *hf_out;

	HANDLE handle_in = GetStdHandle(STD_INPUT_HANDLE);
	hCrt = _open_osfhandle((long) handle_in, _O_TEXT);
	FILE* hf_in = _fdopen(hCrt, "r");
	setvbuf(hf_in, NULL, _IONBF, 128);
	*stdin = *hf_in;
}

void StartHooked(HANDLE& aProcess, HWND& aWindow)
{
	aProcess = 0;
	aWindow = 0;

	PROCESS_INFORMATION pi = {0};
	STARTUPINFO si = {0};
	si.cb = sizeof(si);

	char path[MAX_PATH] = {};
	GetModuleFileName(0, path, sizeof path);
	if (char* end = strrchr(path, '\\'))
		*end = 0;
	strcat_s(path, "\\UIC Hook.dll");

	BOOL ret = DetourCreateProcessWithDll(
		"gw2.exe",
		GetCommandLine(),
		NULL,
		NULL,
		TRUE,
		CREATE_DEFAULT_ERROR_MODE,
		NULL,
		NULL,
		&si,
		&pi,
		NULL,
		path,
		NULL);

	if (!ret)
		throw make_pair("Could not start GW2.exe, please make sure UIC.exe and all related files are placed in the Guild Wars 2 folder.", GetLastError());

	aProcess = pi.hProcess;

	while (!aWindow)
	{
		// wait for application ot finish loading
		WaitForInputIdle(aProcess, INFINITE);

		// find the game window
		aWindow = FindWindowA("ArenaNet_Dx_Window_Class", "Guild Wars 2");

		// no window, wait for one second before trying again
		if (!aWindow)
			if (WaitForSingleObject(aProcess, 1000) == WAIT_OBJECT_0)
			{
				cout << "The application exited before opening a recognized window." << endl;
				throw 0;
			}
	}
}

string FormatError(DWORD aError)
{
	char buf[1024];

	FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		aError,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		buf,
		ARRAYSIZE(buf), NULL);

	return buf;
}

// http://blogs.msdn.com/b/oldnewthing/archive/2005/02/17/375307.aspx#375748
BOOL WaitWithMessageLoop(HANDLE hHandleToWaitOn, DWORD dwInitialTimeOutMilliseconds, DWORD dwIterateTimeOutMilliseconds) 
{ 
	DWORD dwRet=0; 
	MSG msg={0}; 

	dwRet = ::WaitForSingleObject(hHandleToWaitOn, dwInitialTimeOutMilliseconds); 
	if(dwRet == WAIT_OBJECT_0) 
	{ 
		// The object is already signalled. 
		return TRUE; 
	} 

	while(true) 
	{ 
		// There are one or more window message available. Dispatch them. 
		while(::PeekMessage(&msg,NULL,NULL,NULL,PM_REMOVE)) 
		{ 
			::TranslateMessage(&msg); 
			::DispatchMessage(&msg); 
			dwRet = ::WaitForSingleObject(hHandleToWaitOn, 0); 
			if(dwRet == WAIT_OBJECT_0) 
			{ 
				// The object is already signalled. 
				return TRUE; 
			} 
		} 

		// Now we've dispatched all the messages in the queue. 
		// Use MsgWaitForMultipleObjects to either tell us there are 
		// more messages to dispatch, or that our object has been signalled. 

		dwRet = ::MsgWaitForMultipleObjects(1, &hHandleToWaitOn, FALSE, 
			dwIterateTimeOutMilliseconds, QS_ALLINPUT); 

		if(dwRet == WAIT_OBJECT_0) 
		{ 
			// The event was signaled. 
			return TRUE; 
		} 
		else if(dwRet == WAIT_OBJECT_0 + 1) 
		{ 
			// New messages have come that need to be dispatched. 
			continue; 
		} 
		else if(dwRet == WAIT_TIMEOUT) 
		{ 
			// We hit our time limit, continue with the loop. 
			continue; 
		} 
		else 
		{ 
			// Something else happened. 
			return FALSE; 
		} 
	} 

	return FALSE; 
} 
